import pytest
from pyscal_rdf import StructureGraph
from pyscal_rdf.queries import Query

def test_query():
	s = StructureGraph()
	sys = s.create_element("Fe")
	q = Query()

	assert(len(q.sparql.sample_by_latticesystem(s, "Q851536")) == 1)

	struct_gb_1 = s.create_grain_boundary(axis=[0,0,1], 
                        sigma=5, 
                        gb_plane=[3, -1, 0],
                        element='Fe')
	assert(len(q.sparql.sample_by_defect(s, "symmetric tilt")) == 1)
	assert(len(q.sparql.sample_by_sigma(s, 5)) == 1)
	assert(len(q.python.sample_by_altname(s, "bcc")) == 2)







import numpy as np
from json import JSONEncoder

class NumpyArrayEncoder(JSONEncoder):
    """
    Encode numpy to dump in json
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)


import numpy as np
import spglib

def get_angle(vec1, vec2):
    """
    Get angle between two vectors in degrees
    
    Parameters
    ----------
    vec1: list
        first vector
    
    vec2: list
        second vector
    
    Returns
    -------
    angle: float
        angle in degrees
    
    Notes
    -----
    Angle is rounded to two decimal points
    
    """
    return np.round(np.arccos(np.dot(vec1, vec2)/(np.linalg.norm(vec1)*np.linalg.norm(vec2)))*180/np.pi, decimals=2)

def get_coordination(sys):
    sys.find_neighbors(method="cutoff")
    coordination = [len(x) for x in sys.atoms.neighbors.index]
    return coordination

def get_lattice_vector(sys, cartesian=False):
    box = [[sys._structure_dict['scaling_factors'][0], 0, 0], 
           [0, sys._structure_dict['scaling_factors'][1], 0], 
           [0, 0, sys._structure_dict['scaling_factors'][2]]]
    if cartesian:
        box[0][0] = box[0][0]*sys.atoms._lattice_constant 
        box[1][1] = box[1][1]*sys.atoms._lattice_constant 
        box[2][2] = box[2][2]*sys.atoms._lattice_constant
    return box

def get_bravais_lattice(sys):
    lattice = sys.atoms._lattice 
    if lattice == "l12":
        lattice = "https://www.wikidata.org/wiki/Q3006714"
    elif lattice == "b2":
        lattice = "https://www.wikidata.org/wiki/Q851536"
    elif lattice == "diamond":
        lattice = "https://www.wikidata.org/wiki/Q3006714"
    elif lattice == "hcp":
        lattice = "https://www.wikidata.org/wiki/Q663314"
    elif lattice == "a15":
        lattice = "a15"
    elif lattice == "bcc":
        lattice = "https://www.wikidata.org/wiki/Q851536"
    elif lattice == "fcc":
        lattice = "https://www.wikidata.org/wiki/Q3006714"
    return lattice
    
def get_space_group(sys):
    box = get_lattice_vector(sys)
    direct_coordinates = sys._structure_dict['positions']
    atom_types = sys._structure_dict['species']
    results = spglib.get_symmetry_dataset((box,
    direct_coordinates, atom_types))
    return results["international"], results["number"]
    
def get_basis(sys):
    occ_numbers = sys._structure_dict['species']
    tdict = sys.atoms._type_dict
    vals = [val for key, val in tdict.items()]
    
    if vals[0] is not None:
        occ_numbers = [tdict[x] for x in occ_numbers]
    return occ_numbers

    
    




import pytest
from pyscal_rdf import StructureGraph

def test_structuregraph():
	s = StructureGraph()
	sys = s.create_element("Fe")

	vis = s.visualise()
	assert(vis != None)

	vis = s.visualise(styledict={"BNode":{"fontsize":10}})
	assert(vis != None)

	s.write("temp.ttl", format="turtle")
	s = StructureGraph(graph_file="temp.ttl")

	sys = s.create_element("Fe")
	s.add_vacancy(0.5, number=1)

	s = StructureGraph()
	sys = s.create_element("Fe")
	res = s.query_sample("NumberOfAtoms", 2)
	assert(len(res) == 1)


import numpy as np
import pyscal3.core as pc
from rdflib import Graph, Literal, Namespace, XSD, RDF, RDFS, BNode, URIRef, FOAF, SKOS, DCTERMS

CMSO = Namespace("https://purls.helmholtz-metadaten.de/cmso/")
PLDO = Namespace("https://purls.helmholtz-metadaten.de/pldo/")
PODO = Namespace("https://purls.helmholtz-metadaten.de/podo/")

class System(pc.System):
    def __init__(self, filename = None, 
            format = "lammps-dump", 
            compressed = False, 
            customkeys = None,
            source=None):
        super().__init__(filename = filename, 
            format = format, 
            compressed = compressed, 
            customkeys = customkeys)
        #this is the sample which will be stored
        self.sample = None
        #the graph object should also be attached
        #for post-processing of structures
        self.graph = None
        self._atom_ids = None
        if source is not None:
            self.__dict__.update(source.__dict__)

    def __delitem__(self, val):
        if isinstance(val, int):
            val = [val]
        #now the graph has to be updated accordingly
        if self.graph is not None:
            #first annotate graph
            c = (len(val)/self.natoms)
            self.graph.add_vacancy(c, number=len(val))
            #now we need to re-add atoms, so at to remove
            #deleted ones from the vacancy
            atoms = [self._atom_ids[v] for v in val] 
            #this is the list of atoms in this sample
            for atom in atoms:
                #identify the position
                position = list([s[2] for s in self.graph.graph.triples((atom, CMSO.hasPositionVector, None))])[0]
                self.graph.graph.remove((position, None, None))
                #identify element
                element = list([s[2] for s in self.graph.graph.triples((atom, CMSO.hasElement, None))])[0]
                self.graph.graph.remove((element, None, None))
                #now remove the atom from the list completely
                self.graph.graph.remove((atom, None, None))
                self.graph.graph.remove((None, None, atom))
            #now fully remove atoms
            for atom in atoms:
                self._atom_ids.remove(atom)
            #now fix the number of atoms
            self.graph.graph.remove((self.sample, CMSO.hasNumberOfAtoms, None))
            self.graph.graph.add((self.sample, CMSO.hasNumberOfAtoms, Literal(self.natoms-len(val), datatype=XSD.integer)))
            #revamp composition
            #for that first get element
            material = list([s[2] for s in self.graph.graph.triples((self.sample, CMSO.hasMaterial, None))])[0]
            #remove existing chem composution
            self.graph.graph.remove((material, CMSO.hasElementRatio, None))
            #now recalculate and add it again
            chem_comp_element = list(self.composition.keys())
            chem_comp_ratio = [val for key, val in self.composition.items()]
            chem_comp = ["=".join([str(x), str(y)]) for x,y in zip(chem_comp_element, chem_comp_ratio)]
            for x in range(len(chem_comp)):
                self.graph.graph.add((material, CMSO.hasElementRatio, Literal(chem_comp[x], datatype=XSD.string)))
        self.delete(indices=list(val))


import pytest
from pyscal_rdf.encoder import NumpyArrayEncoder
from pyscal_rdf.json_io import write_file
import json
import numpy as np
import os

def test_encoder():
    arr = np.linspace(0, 10, 100)
    data = {"array": arr}
    dumpdata = json.dumps(data, cls=NumpyArrayEncoder)
    assert(np.abs(float(dumpdata.split(',')[-2]) - arr[-2]) < 1E-5)

def test_writer(tmp_path):
    arr = np.linspace(0, 10, 100)
    data = {"array": arr}
    write_file("test_json", data)
    assert(os.path.exists("test_json.json"))


import graphviz
import os
from rdflib import BNode, URIRef, Literal
import uuid
import json
import ipycytoscape

def get_title_from_BNode(x):
    return x.toPython()

def get_string_from_URI(x, ):
    raw = x.toPython().split("#")
    if len(raw)>1:
        return raw[-1]
    else:
        return ".".join(x.toPython().split("/")[-2:])

def parse_object(x):
    if isinstance(x, BNode):
        return get_title_from_BNode(x), "BNode"
    elif isinstance(x, URIRef):
        return get_string_from_URI(x), "URIRef"
    elif isinstance(x, Literal):
        return str(x.title()), "Literal"

styledict = {
    "BNode": {"color": "#ffe6ff", "shape": "box", "style": "filled"},
    "URIRef": {"color": "#ffffcc", "shape": "box", "style": "filled"},
    "Literal": {"color": "#e6ffcc", "shape": "ellipse", "style": "filled"},
}

def _switch_box(box):
    if box == "box":
        return 'rectangle'
    #remember that only boxes will be used, circles no!
    
def _fix_id(string1, istype1):
    if istype1 == 'Literal':
        id1 = str(uuid.uuid4())
    else:
        id1 = string1
    return id1

def visualize_graph(g,
            backend="ipycytoscape",
            edge_color="#37474F",
            styledict=styledict,
            graph_attr ={'rankdir': 'LR'},
            layoutname='cola'):
    if backend=='ipycytoscape':
        return _visualize_graph_ipycytoscape_backend(g,
                                                edge_color=edge_color,
                                                styledict=styledict,
                                                layoutname=layoutname)
    else:
        return _visualize_graph_graphviz_backend(g,
                                                edge_color=edge_color,
                                                styledict=styledict,
                                                graph_attr=graph_attr)
    
def _visualize_graph_ipycytoscape_backend(g,
                                          edge_color="#37474F",
                                          styledict=styledict,
                                          layoutname='cola'):
    #first step is to create the json file
    # we can start with a dict
    gdict = {}
    gdict["nodes"] = []
    gdict["edges"] = []
    for k in g:
        string1, istype1 = parse_object(k[0])
        id1 = _fix_id(string1, istype1)
        d = {"data": {"id": str(id1), 
                      "label": str(string1), 
                      "classes": str(istype1),
                      "shape": _switch_box(styledict[istype1]['shape']),
                      "width": len(string1)*7,
                      "fontsize": styledict[istype1]['fontsize']}}
        gdict["nodes"].append(d)

        string2, istype2 = parse_object(k[2])
        id2 = _fix_id(string2, istype2)
        d = {"data": {"id": str(id2), 
                      "label": str(string2), 
                      "classes": str(istype2),
                      "shape": _switch_box(styledict[istype2]['shape']),
                      "width": len(string2)*7,
                      "fontsize": styledict[istype1]['fontsize']}}
        gdict["nodes"].append(d)
        
        string3, istype3 = parse_object(k[1])
        id3 = str(uuid.uuid4())
        d = {"data": {"id": str(id3), 
                      "label": str(string3), 
                      "source": id1, 
                      "target": id2,
                      "fontsize": styledict[istype1]['fontsize']}}
        gdict["edges"].append(d)
    
    #graph is complete
    #lets try without any customisation
    my_style = [
        {'selector': 'node',
         'style': {
            'font-family': 'arial',
            'font-size': '10px',
            'label': 'data(label)',
            'shape': 'data(shape)',
            "text-valign": "center",
            "text-halign": "center",
            "width": 'data(width)',
            "font-size": 'data(fontsize)',
         }}, 
        {'selector': 
            'edge',
            'style': {
                'font-family': 'arial',
                'label': 'data(label)',
                "font-size": 'data(fontsize)',
                "target-arrow-shape": "triangle",
                "target-arrow-color": edge_color,
                "curve-style": "bezier",
            }},
         {'selector': 
            'node[classes="BNode"]',
            'style': {
                'background-color': styledict['BNode']['color'],
            }}, 
         {'selector': 
            'node[classes="URIRef"]',
            'style': {
                'background-color': styledict['URIRef']['color'],
            }}, 
         {'selector': 
            'node[classes="Literal"]',
            'style': {
                'background-color': styledict['Literal']['color']
            }}, 
    ]
      
    ipycyobj = ipycytoscape.CytoscapeWidget()
    ipycyobj.set_layout(
        name=layoutname,
        avoidOverlap=True,
        animate=True,)
    ipycyobj.graph.add_graph_from_json(gdict, directed=True)
    ipycyobj.set_style(my_style)
    return ipycyobj   
    

def _visualize_graph_graphviz_backend(g,  
            edge_color="#37474F",
            styledict=styledict,
            graph_attr ={'rankdir': 'LR'},
            rankdir='LR'):
    
    dot = graphviz.Digraph()
    for key, val in graph_attr.items():
        dot.graph_attr[key] = val
    for k in g:
        string1, istype1 = parse_object(k[0])
        if istype1 == 'Literal':
            id1 = str(uuid.uuid4())
        else:
            id1 = string1 
        dot.node(id1, label=string1, shape=styledict[istype1]["shape"], 
                 style=styledict[istype1]["style"], 
                 color=styledict[istype1]["color"],
                 fontsize=styledict[istype1]["fontsize"])
        
        string2, istype2 = parse_object(k[2])
        if istype2 == 'Literal':
            id2 = str(uuid.uuid4())
        else:
            id2 = string2 
        dot.node(id2, label=string2, shape=styledict[istype2]["shape"], 
                 style=styledict[istype2]["style"], 
                 color=styledict[istype2]["color"],
                 fontsize=styledict[istype2]["fontsize"])
        
        string3, istype = parse_object(k[1])
        dot.edge(id1, id2, color=edge_color, label=string3, fontsize=styledict[istype2]["fontsize"])
    
    return dot


import pytest
from pyscal_rdf import StructureGraph

def test_structuregraph():
	s = StructureGraph()
	sys = s.create_element("Fe")
	assert(s.sample != None)

	sys = s.create_structure("bcc", element="Fe")
	assert(s.sample != None)

	sys = s.read_structure("tests/al_data/Al.poscar", format="poscar")
	assert(s.sample != None)

	sys = s.create_grain_boundary(axis=[0,0,1], 
                        sigma=5, 
                        gb_plane=[3, -1, 0],
                        element='Fe')
	assert(s.sample != None)







import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

class Network:
    """
    Network representation of CMSO
    """
    def __init__(self):
        self.g = nx.DiGraph()
    
    def add(self, sub, pred, obj, dtype=None, pred_prefix="cmso"):
        pred = f'{pred_prefix}:{pred}'
        self.g.add_node(sub, node_type="object")
        self.g.add_node(pred, node_type="property")
        if dtype is not None:
            nd = "data"
        else:
            nd = "object"
        self.g.add_node(obj, dtype=dtype, node_type=nd)            
        self.g.add_edge(sub, pred)
        self.g.add_edge(pred, obj)
    
    def draw(self):
        nx.draw(self.g, with_labels=True, font_weight='bold')
        
    def get_shortest_path(self, source, target):
        path = nx.shortest_path(self.g, source=source, target=target)
        return path
    
class OntologyNetwork(Network):
    def __init__(self):
        super().__init__()
        self.add("Sample", "hasMaterial", "Material")
        self.add("Material", "hasElementRatio", "ElementRatio", dtype="string")

        self.add("Sample", "hasSimulationCell", "SimulationCell")
        self.add("SimulationCell", "hasVolume", "Volume", dtype="float")
        self.add("Sample", "hasNumberOfAtoms", "NumberOfAtoms", dtype="integer")

        self.add("SimulationCell", "hasLength", "SimulationCellLength")
        self.add("SimulationCellLength", "hasLength_x", "SimulationCellLength_x", dtype="float")
        self.add("SimulationCellLength", "hasLength_y", "SimulationCellLength_y", dtype="float")
        self.add("SimulationCellLength", "hasLength_z", "SimulationCellLength_z", dtype="float")

        self.add("SimulationCell", "hasVector", "SimulationCellVectorA")
        self.add("SimulationCellVectorA", "hasComponent_x", "SimulationCellVectorA_x", dtype="float")
        self.add("SimulationCellVectorA", "hasComponent_y", "SimulationCellVectorA_y", dtype="float")
        self.add("SimulationCellVectorA", "hasComponent_z", "SimulationCellVectorA_z", dtype="float")
        self.add("SimulationCell", "hasVector", "SimulationCellVectorB")
        self.add("SimulationCellVectorB", "hasComponent_x", "SimulationCellVectorB_x", dtype="float")
        self.add("SimulationCellVectorB", "hasComponent_y", "SimulationCellVectorB_y", dtype="float")
        self.add("SimulationCellVectorB", "hasComponent_z", "SimulationCellVectorB_z", dtype="float")
        self.add("SimulationCell", "hasVector", "SimulationCellVectorC")
        self.add("SimulationCellVectorC", "hasComponent_x", "SimulationCellVectorC_x", dtype="float")
        self.add("SimulationCellVectorC", "hasComponent_y", "SimulationCellVectorC_y", dtype="float")
        self.add("SimulationCellVectorC", "hasComponent_z", "SimulationCellVectorC_z", dtype="float")

        self.add("SimulationCell", "hasAngle", "SimulationCellAngle")
        self.add("SimulationCellAngle", "hasAngle_alpha", "SimulationCellAngle_alpha", dtype="float")
        self.add("SimulationCellAngle", "hasAngle_beta", "SimulationCellAngle_beta", dtype="float")
        self.add("SimulationCellAngle", "hasAngle_gamma", "SimulationCellAngle_gamma", dtype="float")

        self.add("Material", "hasStructure", "CrystalStructure")
        self.add("CrystalStructure", "hasAltName", "CrystalStructureAltName", dtype="string")
        self.add("CrystalStructure", "hasSpaceGroupSymbol", "SpaceGroupSymbol", dtype="string")
        self.add("CrystalStructure", "hasSpaceGroupNumber", "SpaceGroupNumber", dtype="integer")

        self.add("CrystalStructure", "hasUnitCell", "UnitCell")
        self.add("UnitCell", "hasBravaisLattice", "LatticeSystem")
        self.add("UnitCell", "hasLatticeParameter", "LatticeParameter")
        self.add("LatticeParameter", "hasLength_x", "LatticeParameter_x", dtype="float")
        self.add("LatticeParameter", "hasLength_y", "LatticeParameter_y", dtype="float")
        self.add("LatticeParameter", "hasLength_z", "LatticeParameter_z", dtype="float")
        self.add("UnitCell", "hasAngle", "LatticeAngle")
        self.add("LatticeAngle", "hasAngle_alpha", "LatticeAngle_alpha", dtype="float")
        self.add("LatticeAngle", "hasAngle_beta", "LatticeAngle_beta", dtype="float")
        self.add("LatticeAngle", "hasAngle_gamma", "LatticeAngle_gamma", dtype="float")

        #add GB properties
        self.add("Material", "hasDefect", "Defect", pred_prefix="cmso")
        self.add("Defect", "type", "GrainBoundary", pred_prefix="rdf")
        self.add("Defect", "type", "TwistBoundary", pred_prefix="rdf")
        self.add("Defect", "type", "TiltBoundary", pred_prefix="rdf")
        self.add("Defect", "type", "SymmetricTiltBoundary", pred_prefix="rdf")
        self.add("Defect", "type", "MixedBoundary", pred_prefix="rdf")
        self.add("Defect", "hasSigmaValue", "Sigma", dtype="integer", pred_prefix="pldo")
        self.add("Defect", "hasGBPlane", "GBPlane", pred_prefix="pldo", dtype="string")
        self.add("Defect", "hasRotationAxis", "RotationAxis", pred_prefix="pldo", dtype="string")
        self.add("Defect", "hasMisorientationAngle", "MisorientationAngle", pred_prefix="pldo", dtype="float")

        #add vacancy
        self.add("Defect", "type", "Vacancy", pred_prefix="rdf")
        self.add("SimulationCell", "hasVacancyConcentration", "VacancyConcentration", pred_prefix="podo", dtype="float")
        self.add("SimulationCell", "hasNumberOfVacancies", "NumberOfVacancy", pred_prefix="podo", dtype="integer")
         

    def get_path_from_sample(self, target):
        path = self.get_shortest_path(source="Sample", target=target)
        triplets = []
        for x in range(len(path)//2):
            triplets.append(path[2*x:2*x+3])
        return triplets
        
    def formulate_query(self, target, value):
        #first get triplets
        triplets = self.get_path_from_sample(target)
        #start building query
        query = self._formulate_query_path(triplets)
        query.append(self._formulate_filter_expression(triplets, value))
        query.append("}")
        query = " ".join(query)
        return query
        
    
    def _formulate_query_path(self, triplets):
        query = []
        query.append("PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>")
        query.append("PREFIX pldo: <https://purls.helmholtz-metadaten.de/pldo/>")
        query.append("PREFIX podo: <https://purls.helmholtz-metadaten.de/podo/>")
        query.append("PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>")
        query.append("SELECT DISTINCT ?sample")
        query.append("WHERE {")
        for triple in triplets:
            query.append("    ?%s %s ?%s ."%(triple[0].lower(), 
                                                  triple[1], 
                                                  triple[2].lower()))
        return query
    
    def _formulate_filter_expression(self, triplets, value):                       
        value, datatype = self._check_value(value)      
        last_val = self.g.nodes[triplets[-1][-1]]
        last_val_name = triplets[-1][-1].lower()
        
        #if it is nodetype data
        if last_val['node_type'] == "data":
            if datatype == "multi_string":
                qstr = self._formulate_or_string_query(last_val, 
                                                   last_val_name, 
                                                   value)
            elif datatype == "multi_number":
                qstr = self._formulate_range_number_query(last_val, 
                                                   last_val_name, 
                                                   value)
            else:
                qstr = self._formulate_equal_query(last_val, 
                                                   last_val_name, 
                                                   value)
            return qstr
        else:
            raise NotImplementedError("Non-data queries are not implemented")
    
    def _check_value(self, value):
        if isinstance(value, list):
            if not len(value) == 2:
                raise ValueError("value can be maximum length 2")
        else:
            value = [value]
        if all(isinstance(x, str) for x in value):
            datatype = "string"
        elif all(isinstance(x, (int, float)) for x in value):
            datatype = "number"
        else:
            raise TypeError("Values have to be of same type")
        if len(value) == 1:
            datatype = f'single_{datatype}'
        else:
            datatype = f'multi_{datatype}'
        return value, datatype
    
    
    def _formulate_equal_query(self, last_val, last_val_name, value):
        qstr = "FILTER (?%s=\"%s\"^^xsd:%s)"%(last_val_name, 
                                              str(value[0]), 
                                              last_val['dtype'])
        return qstr
    
    def _formulate_or_string_query(self, last_val, last_val_name, value):
        qstr = "FILTER (?%s=\"%s\"^^xsd:%s || ?%s=\"%s\"^^xsd:%s)"%(last_val_name, 
                                                                    str(value[0]), 
                                                                    last_val['dtype'],
                                                                    last_val_name, 
                                                                    str(value[1]), 
                                                                    last_val['dtype'],)
        return qstr
    
    def _formulate_range_number_query(self, last_val, last_val_name, value):
        value = np.sort(value)
        qstr = "FILTER (?%s >= \"%s\"^^xsd:%s && ?%s <= \"%s\"^^xsd:%s)"%(last_val_name, 
                                                                    str(value[0]), 
                                                                    last_val['dtype'],
                                                                    last_val_name, 
                                                                    str(value[1]), 
                                                                    last_val['dtype'],)
        return qstr

        
            
            
            

        
        


"""
StructureGraph is the central object in pyscal_rdf which combines all the functionality
of :py:class:`pyscal_rdf.graph.RDFGraph` along with easy structural creation routines.
"""
import numpy as np
import copy

from pyscal_rdf.rdfsystem import System
from pyscal3.crystal_structures import structure_creator, elements, structures
from pyscal_rdf.graph import RDFGraph
from pyscal3.grain_boundary import GrainBoundary

class StructureGraph(RDFGraph):
    def __init__(self, graph_file=None, 
        store="Memory", 
        store_file=None,
        identifier="default_graph"):
        
        super().__init__(graph_file=graph_file, store=store, store_file=store_file, identifier=identifier)
        self._element_dict = elements
        self._structure_dict = structures

    def create_element(self, element, repetitions=(1,1,1), 
                       noise=0, add_to_graph=True, names=True):
        """
        Create a crystal structure of the given element

        Parameters
        ----------
        element : string
            chemical symbol of the element

        repetitions : list of ints of len 3, optional
            of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
            default `[1, 1, 1]`.

        noise : float, optional
            If provided add normally distributed noise with standard deviation `noise` to the atomic positions.
        
        add_to_graph: bool, optinal
            If False, the created structure will not be added to the graph

        names: bool, optional
            If True, names will be used as IDs

        Returns
        -------
        System: :py:class:`pyscal.core.System`

        Notes
        -----
        Python module Mendelev is used to get the lattice constant with which the system will be constructed.
        If it is an hexagonal lattice, the ideal c/a ration will be used.

        """
        if element in self._element_dict.keys():
            structure = self._element_dict[element]['structure']
            sys = System(source=structure_creator(structure,
                        repetitions=repetitions,
                        noise=noise,
                        lattice_constant=self._element_dict[element]['lattice_constant'],
                        element = element))
            if add_to_graph:
                self.add_structure_to_graph(sys, names=names)
                #sys.sample = self.sample
                #sys._atom_ids = copy.copy(self._atom_ids)
                #sys.graph = self
            return sys
    
    def create_structure(self, structure, 
                         lattice_constant = 1.00, 
                         repetitions = None, ca_ratio = 1.633, 
                         noise = 0, element=None,
                         add_to_graph=True, names=True):
        """
        Create a crystal structure and return it as a System object.

        Parameters
        ----------
        structure : {'sc', 'bcc', 'fcc', 'hcp', 'diamond', 'a15' or 'l12'}
            type of the crystal structure

        lattice_constant : float, optional
            lattice constant of the crystal structure, default 1

        repetitions : list of ints of len 3, optional
            of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
            default `[1, 1, 1]`.

        ca_ratio : float, optional
            ratio of c/a for hcp structures, default 1.633

        noise : float, optional
            If provided add normally distributed noise with standard deviation `noise` to the atomic positions.

        add_to_graph: bool, optinal
            If False, the created structure will not be added to the graph

        names: bool, optional
            If True, names will be used as IDs
        
        element : string, optional
            The chemical element

        Returns
        -------
        System: pyscal System

        """
        if structure in self._structure_dict.keys():
            sys = System(source=structure_creator(structure,
                        repetitions=repetitions,
                        noise=noise,
                        lattice_constant=lattice_constant,
                        element = element,
                        ))
            if add_to_graph:
                self.add_structure_to_graph(sys, names = names)
                #sys.sample = self.sample
                #sys._atom_ids = copy.copy(self._atom_ids)
                #sys.graph = self
            return sys
    
    def read_structure(self, filename, format="lammps-dump",
                      add_to_graph=True, names=True):
        """
        Read an input file and return it as a System object.

        Parameters
        ----------
        filename: string
            name of the input file

        format: string
            format of the input file

        add_to_graph: bool, optinal
            If False, the created structure will not be added to the graph

        names: bool, optional
            If True, names will be used as IDs

        Returns
        -------
        System: pyscal System
        system will be populated with given atoms and simulation box

        """
        sys = System(filename, format=format)
        if add_to_graph:
            self.add_structure_to_graph(sys, names=names)
            #sys.sample = self.sample
            #sys._atom_ids = copy.copy(self._atom_ids)
            #sys.graph = self
        return sys
    
    def create_grain_boundary(self, axis, 
                              sigma, gb_plane,
                              structure=None,
                              element=None, 
                              lattice_constant=1,
                              repetitions=(1,1,1),
                              overlap=0.0,
                              add_to_graph=True,
                              names=True):
        """
        Create a grain boundary structure and return it as a System object.

        Parameters
        ----------
        axis: list of ints of length 3
            The grain boundary axis

        sigma: int
            sigma value of the grain boundary
        
        gb_plane: list of ints of length 3
            The grain boundary plane

        structure : {'sc', 'bcc', 'fcc', 'hcp', 'diamond', 'a15' or 'l12'}
            type of the crystal structure

        element : string, optional
            The chemical element

        lattice_constant : float, optional
            lattice constant of the crystal structure, default 1

        repetitions : list of ints of len 3, optional
            of type `[nx, ny, nz]`, repetions of the unit cell in x, y and z directions.
            default `[1, 1, 1]`.

        overlap: float, optional
            overlap between the two grains

        add_to_graph: bool, optinal
            If False, the created structure will not be added to the graph

        names: bool, optional
            If True, names will be used as IDs            

        Returns
        -------
        System: pyscal System

        """
        gb = GrainBoundary()
        gb.create_grain_boundary(axis=axis, sigma=sigma, 
                                 gb_plane=gb_plane)

        #use standard creation routine
        if structure is not None:
            sys = System(source=gb.populate_grain_boundary(structure, 
                                             repetitions = repetitions,
                                             lattice_parameter = lattice_constant,
                                             overlap=overlap))
        elif element is not None:
            sys = System(source=gb.populate_grain_boundary(element, 
                                             repetitions=repetitions,
                                             overlap=overlap))
        else:
            raise ValueError("Either structure or element should be provided")
            
        #mapping of the system can be done
        self.add_structure_to_graph(sys, names=names)
        #sys.sample = self.sample
        #sys._atom_ids = copy.copy(self._atom_ids)
        #sys.graph = self
        gb_dict = {"GBPlane": " ".join(np.array(gb_plane).astype(str)),
                  "RotationAxis": axis,
                  "MisorientationAngle": gb.theta,
                  "GBType": gb.find_gb_character(),
                  "sigma": gb.sigma,
                  }
        self.add_gb(gb_dict)
        return sys
            




from setuptools import setup, find_packages


with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='pyscal_rdf',
    version='0.0.13',
    author='Abril Azocar Guzman, Sarath Menon',
    author_email='sarath.menon@pyscal.org',
    description='Ontology based structural manipulation and quering',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(include=['pyscal_rdf', 'pyscal_rdf.*']),
    zip_safe=False,
    download_url = 'https://github.com/pyscal/pyscal_rdf',
    url = 'https://pyscal.org',
    install_requires=['numpy', 'ase', 'rdflib', 
    'pyyaml', 'graphviz', 'networkx', 
    'ipycytoscape', 'pyscal3', 'spglib',],
    classifiers=[
        'Programming Language :: Python :: 3'
    ],
    include_package_data=True,
)



import json
import yaml
from pyscal_rdf.encoder import NumpyArrayEncoder

def write_file(outfile, data):
    """
    Write a given dict as json file
    
    Parameters
    ----------
    outfile: string
        name of output file. `.json` will be added to the given file name
    
    data: dict
        input data dict
    
    Returns
    -------
    None
    """
    with open(".".join([outfile, "json"]), "w") as fout:
        json.dump(data, fout, cls=NumpyArrayEncoder)
    #with open(".".join([outfile, "yaml"]), "w") as fout:
    #    yaml.unsafe_dump(convert_to_dict(sys), fout)


from rdflib import URIRef, Namespace, BNode, Literal, XSD
from rdflib.namespace import RDF
import numpy as np

CMSO = Namespace("https://purls.helmholtz-metadaten.de/cmso/")

class Query:
    def __init__(self):
        self.sparql = Sparql()
        self.python = Python()

class Sparql:
    def __init__(self):
        pass
    
    def sample_by_latticesystem(self, g, latticesystem):
        query = """
        PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>
        PREFIX wiki: <https://www.wikidata.org/wiki/>
        SELECT DISTINCT ?sample
        WHERE {
            ?unitcell cmso:hasBravaisLattice wiki:%s .
            ?structure cmso:hasUnitCell ?unitcell .
            ?material cmso:hasStructure ?structure .
            ?sample cmso:hasMaterial ?material .
        }"""%(latticesystem)
        qres = g.graph.query(query)
        samples = [j[0] for j in qres]
        return samples
    
    def sample_by_sigma(self, g, sigma):
        if isinstance(sigma, int):
            query="""
            PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>
            PREFIX pldo: <https://purls.helmholtz-metadaten.de/pldo/>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            SELECT DISTINCT ?sample
            WHERE {
                ?sample cmso:hasMaterial ?material .
                ?material cmso:hasDefect ?defect .
                ?defect pldo:hasSigmaValue ?sigma .
            FILTER (?sigma="%d"^^xsd:integer)
            }"""%(sigma)
        elif isinstance(sigma, list):
            if not len(sigma) == 2:
                raise ValueError("Range queries support only arrays of length 2")
            sigma = np.sort(np.abs(sigma))
            query="""
            PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>
            PREFIX pldo: <https://purls.helmholtz-metadaten.de/pldo/>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            SELECT DISTINCT ?sample
            WHERE {
                ?sample cmso:hasMaterial ?material .
                ?material cmso:hasDefect ?defect .
                ?defect pldo:hasSigmaValue ?sigma .
            FILTER (?sigma >= "%d"^^xsd:integer && ?sigma <= "%d"^^xsd:integer)
            }"""%(sigma[0], sigma[1])
        qres = g.graph.query(query)
        samples = [j[0] for j in qres]
        return samples          
            
    def sample_by_defect(self, g, defect):
        defect = defect.lower()
        defectdict = {"tilt": "TiltBoundary", 
                      "twist": "TwistBoundary", 
                      "symmetric tilt": "SymmetricTiltBoundary", 
                      "mixed": "MixedBoundary", 
                      "grain boundary": None}
        
        if defect not in defectdict.keys():
            raise ValueError(f'Please choose defect from {defectlist.keys()}')
        if defect != "grain boundary":
            query = """
            PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>
            PREFIX pldo: <https://purls.helmholtz-metadaten.de/pldo/>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            SELECT DISTINCT ?sample
            WHERE {
                ?sample cmso:hasMaterial ?material .
                ?material cmso:hasDefect ?defect .
                ?defect rdf:type pldo:%s .
            }"""%(defectdict[defect])
        else:
            querycomm = ['{?defect rdf:type pldo:%s}'%val for key, val in defectdict.items()]
            querycomm = " UNION ".join(querycomm)
            query = """
            PREFIX cmso: <https://purls.helmholtz-metadaten.de/cmso/>
            PREFIX pldo: <https://purls.helmholtz-metadaten.de/pldo/>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            SELECT DISTINCT ?sample
            WHERE {
                ?sample cmso:hasMaterial ?material .
                ?material cmso:hasDefect ?defect .
                %s .
            }"""%(querycomm)
        qres = g.graph.query(query)
        samples = [j[0] for j in qres]
        return samples        
class Python:
    def __init__(self):
        pass
    
    def sample_by_altname(self, g, altname):
        samples = []
        for s in g.graph.triples((None, CMSO.hasAltName, Literal(altname, datatype=XSD.string))):
            for k in g.graph.triples((None, CMSO.hasStructure, s[0])):
                for j in g.graph.triples((None, CMSO.hasMaterial, k[0])):
                    samples.append(j[0])
        return samples
                
            
        

    



from pyscal_rdf import StructureGraph
import pytest


def test_db():
	g = StructureGraph(store="SQLAlchemy", store_file="testfile.db")
	struct_Fe = g.create_element("Fe")
	assert g.n_samples == 1

	k = StructureGraph(store="SQLAlchemy", store_file="testfile.db")
	assert k.n_samples == 1



"""
Graph module contains the basic RDFGraph object in pyscal_rdf. This object gets a structure
as an input and annotates it with the CMSO ontology (PLDO and PODO too as needed). The annotated
object is stored in triplets.
"""

from rdflib import Graph, Literal, Namespace, XSD, RDF, RDFS, BNode, URIRef, FOAF, SKOS, DCTERMS
from rdflib.store import NO_STORE, VALID_STORE

import os
import numpy as np
import inspect
from ase.io import write
import copy

from pyscal_rdf.visualize import visualize_graph
from pyscal_rdf.network import OntologyNetwork
from pyscal_rdf.rdfsystem import System
import pyscal_rdf.properties as prp
#from pyscal3.core import System
from pyscal3.atoms import Atoms

CMSO = Namespace("https://purls.helmholtz-metadaten.de/cmso/")
PLDO = Namespace("https://purls.helmholtz-metadaten.de/pldo/")
PODO = Namespace("https://purls.helmholtz-metadaten.de/podo/")

defstyledict = {
    "BNode": {"color": "#ffe6ff", 
              "shape": "box", 
              "style": "filled",
              "fontsize": "8",
              "fontname": "Courier"},
    "URIRef": {"color": "#ffffcc", 
               "shape": "box", 
               "style": "filled",
               "fontsize": "8"},
    "Literal": {"color": "#e6ffcc", 
                "shape": "parallelogram", 
                "style": "filled",
                "fontsize": "8"},
}

def _replace_keys(refdict, indict):
    for key, val in indict.items():
        if key in refdict.keys():
            if isinstance(val, dict):
                _replace_keys(refdict[key], indict[key])
            else:
                refdict[key] = val
    return refdict

class RDFGraph:
    def __init__(self, graph_file=None, 
        store="Memory", 
        store_file=None,
        identifier="default_graph"):
        
        #owlfile = os.path.join(os.path.dirname(__file__), "data/cmso.owl")
        #self.graph.parse(owlfile, format='xml')
        if store == "Memory":
            self.graph = Graph(store="Memory", identifier=identifier)

        elif store=="SQLAlchemy":
            if store_file is None:
                raise ValueError("store file is needed if store is not memory")
            self.graph = Graph(store="SQLAlchemy", identifier=identifier)
            uri = Literal(f"sqlite:///{store_file}")
            self.graph.open(uri, create=True)

        elif inspect.isclass(type(store)):
            try:
                prpath = store.path
                dbfile = os.path.join(prpath, 'project.db')
                #now start sqlalchemy instance
                self.graph = Graph(store="SQLAlchemy", identifier=identifier)
                uri = Literal(f"sqlite:///{dbfile}")
                self.graph.open(uri, create=True)
            except:
                raise ValueError("store should be pyiron_project, SQLAlchemy, or Memory")
        
        else:
            raise ValueError("store should be pyiron_project, SQLAlchemy, or Memory")

        self.graph.bind("cmso", CMSO)
        self.graph.bind("pldo", PLDO)
        
        if graph_file is not None:
            if os.path.exists(graph_file):
                self.graph.parse(graph_file)
        
        self.sample = None
        self.material = None
        self.sysdict = None
        self.sgraph = None
        self._query_graph = OntologyNetwork()
        self._atom_ids = None
    
    def process_structure(self, structure, format=None):
        """
        Convert a given :py:class:`pyscal.core.System` to a data dictionary which can be used for annotation
        and storing the data in the RDF Graph.

        Parameters
        ----------
        structure: :py:class:`pyscal.core.System`
            input structure

        Returns
        -------
        None
        """
        if isinstance(structure, System):
            #self.sysdict = convert_to_dict(structure)
            self.sys = structure
        elif os.path.exists(structure):
            sys = System(structure, format=format)
            #self.sysdict = convert_to_dict(sys)
            self.sys = sys
    
    def data(self, key):
        #this method gets info directly from the dict
        if key=="ChemicalCompositionElement":
            return list(self.sys.composition.keys())
        elif key=="ChemicalCompositionRatio":
            return [val for key, val in self.sys.composition.items()]
        elif key=="CellVolume":
            return self.sys.volume
        elif key=="NumberOfAtoms":
            return self.sys.natoms
        elif key=="SimulationCellLengthX":
            return self.sys.box_dimensions[0]
        elif key=="SimulationCellLengthY":
            return self.sys.box_dimensions[1]
        elif key=="SimulationCellLengthZ":
            return self.sys.box_dimensions[2]

        elif key=="SimulationCellVectorA":
            return self.sys.box[0]
        elif key=="SimulationCellVectorB":
            return self.sys.box[1]
        elif key=="SimulationCellVectorC":
            return self.sys.box[2]
        
        elif key=="SimulationCellAngleAlpha":
            return prp.get_angle(self.sys.box[0], self.sys.box[1])
        elif key=="SimulationCellAngleBeta":
            return prp.get_angle(self.sys.box[1], self.sys.box[2])
        elif key=="SimulationCellAngleGamma": 
            return prp.get_angle(self.sys.box[2], self.sys.box[0])
        
        elif key=="Element":
            if self.sys.atoms.species[0] is not None:
                return self.sys.atoms.species
            else:
                return self.sys.atoms.types
        elif key=="Coordination":
            return prp.get_coordination(self.sys)
        elif key=="Positions":
            return self.sys.atoms.positions
        elif key=="LatticeParameter":
            return self.sys.atoms._lattice_constant      
        
        elif key=="SpaceGroupSymbol":
            if self.sys._structure_dict is not None:
                symbol, number = prp.get_space_group(self.sys)
                return symbol
        elif key=="SpaceGroupNumber":
            if self.sys._structure_dict is not None:
                symbol, number = prp.get_space_group(self.sys)
                return number
            else:
                return None

        elif key=="CrystalStructureName":
            if self.sys._structure_dict is not None:
                return self.sys.atoms._lattice 
            else:
                return None

        elif key=="BravaisLattice":
            if self.sys._structure_dict is not None:
                return prp.get_bravais_lattice(self.sys)
            else:
                return None

        elif key=="BasisPositions":
            if self.sys._structure_dict is not None:
                return self.sys._structure_dict['positions']
            else:
                return None

        elif key=="BasisOccupancy":
            if self.sys._structure_dict is not None:
                return prp.get_basis(self.sys)
            else:
                return None

        elif key=="LatticeVectors":
            if self.sys._structure_dict is not None:
                return prp.get_lattice_vector(self.sys)
            else:
                return None

        #if self.sysdict is not None:
        #    if key in self.sysdict:
        #        return self.sysdict[key]
        return None
    
    def add(self, triple):
        if str(triple[2].toPython()) != 'None':
            self.graph.add(triple)
        
    def add_structure_to_graph(self, structure, names=True, name_index=None, format=None):
        """
        Add a given :py:class:`pyscal.core.System` to the Graph object

        Parameters
        ----------
        structure: :py:class:`pyscal.core.System`
            input structure

        names: bool
            if True, alphanumeric names will be used instead of random BNodes

        Returns
        -------
        None
        """
        self.process_structure(structure, format=format)
        #now add to graph
        if name_index is None:
            name_index = self.n_samples + 1
        self.create_graph(names=names, name_index=name_index)
        structure.sample = self.sample
        structure._atom_ids = copy.copy(self._atom_ids)
        structure.graph = self
    
    def create_graph(self, names=False, name_index="1"):
        """
        Create the RDF Graph from the data stored

        Parameters
        ----------
        names: bool
            if True, alphanumeric names will be used instead of random BNodes

        name_index: string
            Prefix to be added to identifiers, default 01        
        
        Returns
        -------
        None
        """

        if names:
            name_list = [f'Sample_{name_index}', f'Material_{name_index}',
                        f'ChemicalComposition_{name_index}', f'SimulationCell_{name_index}',
                        f'SimulationCell_{name_index}', f'CrystalStructure_{name_index}',
                        f'SpaceGroup_{name_index}', f'UnitCell_{name_index}',
                        f'UnitCell_{name_index}', f'Atom_{name_index}']
        else:
            name_list = [None, None,
                        None, None,
                        None, None,
                        None, None,
                        None, None]
        self.add_sample(name=name_list[0])
        self.add_material(name=name_list[1])
        self.add_chemical_composition(name=name_list[2])
        self.add_simulation_cell(name=name_list[3])
        self.add_simulation_cell_properties(name=name_list[4])
        self.add_crystal_structure(name=name_list[5])
        self.add_space_group(name=name_list[6])
        self.add_unit_cell(name=name_list[7])
        self.add_lattice_properties(name=name_list[8])
        self.add_atoms(name=name_list[9])

        self.add((CMSO.SimulationCellLength, RDFS.subClassOf, CMSO.Length))
        self.add((CMSO.LatticeParameter, RDFS.subClassOf, CMSO.Length))
        self.add((CMSO.Length, CMSO.hasUnit, URIRef("https://qudt.org/2.1/vocab/unit#ANGSTROM")))
        
        self.add((CMSO.SimulationCellAngle, RDFS.subClassOf, CMSO.Angle))
        self.add((CMSO.LatticeAngle, RDFS.subClassOf, CMSO.Angle))
        self.add((CMSO.Angle, CMSO.hasUnit, URIRef("https://qudt.org/2.1/vocab/unit#DEG")))
        
        self.add((CMSO.LatticeVector, RDFS.subClassOf, CMSO.Vector))
        self.add((CMSO.SimulationCellVector, RDFS.subClassOf, CMSO.Vector))
        self.add((CMSO.PositionVector, RDFS.subClassOf, CMSO.Vector))
        self.add((CMSO.Vector, CMSO.hasUnit, URIRef("https://qudt.org/2.1/vocab/unit#ANGSTROM")))
        
        
    def add_sample(self, name=None):
        """
        Add a CMSO Sample object

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        sample_01 = BNode(name)
        self.add((sample_01, RDF.type, CMSO.AtomicScaleSample))
        self.sample = sample_01
    
    def add_material(self, name=None):
        """
        Add a CMSO Material object

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        material_01 = BNode(name)
        self.add((self.sample, CMSO.hasMaterial, material_01))
        self.add((material_01, RDF.type, CMSO.CrystallineMaterial))        
        self.material = material_01
    
    def add_chemical_composition(self, name=None):
        """
        Add chemical composition

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        chem_comp = ["=".join([str(x), str(y)]) for x,y in zip(self.data("ChemicalCompositionElement"), self.data("ChemicalCompositionRatio"))]
        for x in range(len(chem_comp)):
            self.add((self.material, CMSO.hasElementRatio, Literal(chem_comp[x], datatype=XSD.string)))
    
    def add_simulation_cell(self, name=None):
        """
        Add a CMSO SimulationCell

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        simulation_cell_01 = BNode(name)
        self.add((self.sample, CMSO.hasSimulationCell, simulation_cell_01))
        self.add((simulation_cell_01, RDF.type, CMSO.SimulationCell))
        self.add((simulation_cell_01, CMSO.hasVolume, Literal(np.round(self.data("CellVolume"), decimals=2), datatype=XSD.float)))
        self.add((self.sample, CMSO.hasNumberOfAtoms, Literal(self.data("NumberOfAtoms"), datatype=XSD.integer)))
        self.simulation_cell = simulation_cell_01
        
    
    def add_simulation_cell_properties(self, name=None):
        """
        Add a CMSO SimulationCell properties such as SimulationCellLength,
        and Vectors.

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        uname = None
        if name is not None:
            uname = f'{name}Length'
        simulation_cell_length_01 = BNode(uname)
        self.add((self.simulation_cell, CMSO.hasLength, simulation_cell_length_01))
        self.add((simulation_cell_length_01, RDF.type, CMSO.SimulationCellLength))
        self.add((simulation_cell_length_01, CMSO.hasLength_x, Literal(self.data("SimulationCellLengthX"), datatype=XSD.float)))
        self.add((simulation_cell_length_01, CMSO.hasLength_y, Literal(self.data("SimulationCellLengthY"), datatype=XSD.float)))
        self.add((simulation_cell_length_01, CMSO.hasLength_z, Literal(self.data("SimulationCellLengthZ"), datatype=XSD.float)))
        
        uname = None
        if name is not None:
            uname = f'{name}Vector01'
        simulation_cell_vector_01 = BNode(uname)
        self.add((self.simulation_cell, CMSO.hasVector, simulation_cell_vector_01))
        self.add((simulation_cell_vector_01, RDF.type, CMSO.SimulationCellVector))
        self.add((simulation_cell_vector_01, CMSO.hasComponent_x, Literal(self.data("SimulationCellVectorA")[0], datatype=XSD.float)))
        self.add((simulation_cell_vector_01, CMSO.hasComponent_y, Literal(self.data("SimulationCellVectorA")[1], datatype=XSD.float)))
        self.add((simulation_cell_vector_01, CMSO.hasComponent_z, Literal(self.data("SimulationCellVectorA")[2], datatype=XSD.float)))
        
        uname = None
        if name is not None:
            uname = f'{name}Vector02'
        simulation_cell_vector_02 = BNode(uname)
        self.add((self.simulation_cell, CMSO.hasVector, simulation_cell_vector_02))
        self.add((simulation_cell_vector_02, RDF.type, CMSO.SimulationCellVector))
        self.add((simulation_cell_vector_02, CMSO.hasComponent_x, Literal(self.data("SimulationCellVectorB")[0], datatype=XSD.float)))
        self.add((simulation_cell_vector_02, CMSO.hasComponent_y, Literal(self.data("SimulationCellVectorB")[1], datatype=XSD.float)))
        self.add((simulation_cell_vector_02, CMSO.hasComponent_z, Literal(self.data("SimulationCellVectorB")[2], datatype=XSD.float)))
        
        uname = None
        if name is not None:
            uname = f'{name}Vector03'
        simulation_cell_vector_03 = BNode(uname)
        self.add((self.simulation_cell, CMSO.hasVector, simulation_cell_vector_03))
        self.add((simulation_cell_vector_03, RDF.type, CMSO.SimulationCellVector))
        self.add((simulation_cell_vector_03, CMSO.hasComponent_x, Literal(self.data("SimulationCellVectorC")[0], datatype=XSD.float)))
        self.add((simulation_cell_vector_03, CMSO.hasComponent_y, Literal(self.data("SimulationCellVectorC")[1], datatype=XSD.float)))
        self.add((simulation_cell_vector_03, CMSO.hasComponent_z, Literal(self.data("SimulationCellVectorC")[2], datatype=XSD.float)))
        
        uname = None
        if name is not None:
            uname = f'{name}Angle'
        simulation_cell_angle_01 = BNode(uname)
        self.add((self.simulation_cell, CMSO.hasAngle, simulation_cell_angle_01))
        self.add((simulation_cell_angle_01, RDF.type, CMSO.SimulationCellAngle))
        self.add((simulation_cell_angle_01, CMSO.hasAngle_alpha, Literal(self.data("SimulationCellAngleAlpha"), datatype=XSD.float)))
        self.add((simulation_cell_angle_01, CMSO.hasAngle_beta, Literal(self.data("SimulationCellAngleBeta"), datatype=XSD.float)))
        self.add((simulation_cell_angle_01, CMSO.hasAngle_gamma, Literal(self.data("SimulationCellAngleGamma"), datatype=XSD.float)))
        
    
    def add_crystal_structure(self, name=None):
        """
        Add a CMSO Crystal Structure

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        crystal_structure_01 = BNode(name)
        self.add((self.material, CMSO.hasStructure, crystal_structure_01))
        self.add((crystal_structure_01, RDF.type, CMSO.CrystalStructure))    
        self.add((crystal_structure_01, CMSO.hasAltName, Literal(self.data("CrystalStructureName"), datatype=XSD.string)))
        self.crystal_structure = crystal_structure_01
        
    def add_space_group(self, name=None):
        """
        Add a CMSO Space Group

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """
        self.add((self.crystal_structure, CMSO.hasSpaceGroupSymbol, Literal(self.data("SpaceGroupSymbol"), datatype=XSD.string)))
        self.add((self.crystal_structure, CMSO.hasSpaceGroupNumber, Literal(self.data("SpaceGroupNumber"), datatype=XSD.integer)))
    
            
    def add_unit_cell(self, name=None):
        """
        Add a CMSO Unit Cell

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        unit_cell_01 = BNode(name)
        self.add((self.crystal_structure, CMSO.hasUnitCell, unit_cell_01))
        self.add((unit_cell_01, RDF.type, CMSO.UnitCell))
        self.unit_cell = unit_cell_01
        
        #add bravais lattice
        bv = None
        if self.data("BravaisLattice") is not None:
            bv = URIRef(self.data("BravaisLattice"))
            self.add((self.unit_cell, CMSO.hasBravaisLattice, bv))
        
    def add_lattice_properties(self, name=None):
        """
        Add CMSO lattice properties such as Lattice Parameter,
        and its lengths and angles. 

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        uname = None
        if name is not None:
            uname = f'{name}LatticeParameter'
        lattice_parameter_01 = BNode(uname)
        self.add((self.unit_cell, CMSO.hasLatticeParamter, lattice_parameter_01))
        self.add((lattice_parameter_01, RDF.type, CMSO.LatticeParameter))
        self.add((lattice_parameter_01, CMSO.hasLength_x, Literal(self.data("LatticeParameter"), datatype=XSD.float)))
        self.add((lattice_parameter_01, CMSO.hasLength_y, Literal(self.data("LatticeParameter"), datatype=XSD.float)))
        self.add((lattice_parameter_01, CMSO.hasLength_z, Literal(self.data("LatticeParameter"), datatype=XSD.float)))
        
        uname = None
        if name is not None:
            uname = f'{name}LatticeAngle'
        lattice_angle_01 = BNode(uname)
        self.add((self.unit_cell, CMSO.hasAngle, lattice_angle_01))
        self.add((lattice_angle_01, RDF.type, CMSO.LatticeAngle))
        self.add((lattice_angle_01, CMSO.hasAngle_alpha, Literal(90, datatype=XSD.float)))
        self.add((lattice_angle_01, CMSO.hasAngle_beta, Literal(90, datatype=XSD.float)))
        self.add((lattice_angle_01, CMSO.hasAngle_gamma, Literal(90, datatype=XSD.float)))        
        
    def add_atoms(self, name=None):
        """
        Add Atoms including their species and positions

        Parameters
        ----------
        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """
        self._atom_ids = []

        positions = self.data("Positions")
        elements = self.data("Element")
        coordination = self.data("Coordination")
        for x in range(len(positions)):
            uname = None
            if name is not None:
                uname = f'{name}_{x}'            
            #create atom
            atom = BNode(uname)
            self._atom_ids.append(atom)
            self.add((self.sample, CMSO.hasAtom, atom))
            self.add((atom, RDF.type, CMSO.Atom))

            uname = None
            if name is not None:
                uname = f'{name}_{x}_Position'            
            position = BNode(uname)
            self.add((atom, CMSO.hasPositionVector, position))
            self.add((position, RDF.type, CMSO.PositionVector))
            self.add((position, CMSO.hasComponent_x, Literal(positions[x][0],
                                                                  datatype=XSD.float)))
            self.add((position, CMSO.hasComponent_y, Literal(positions[x][1],
                                                                  datatype=XSD.float)))
            self.add((position, CMSO.hasComponent_z, Literal(positions[x][2],
                                                                  datatype=XSD.float)))
            #now add coordination
            uname = None
            if name is not None:
                uname = f'{name}_{x}_Element'            
            element = BNode(uname)
            self.add((atom, CMSO.hasElement, element))
            self.add((element, RDF.type, CMSO.Element))
            self.add((element, CMSO.hasSymbol, Literal(str(elements[x]),
                                                            datatype=XSD.string)))
            #finally occupancy
            self.add((atom, CMSO.hasCoordinationNumber, Literal(coordination[x],
                                                                     datatype=XSD.integer)))

    
    
    def add_gb(self, gb_dict, name=None):
        """
        Add GB details which will be annotated using PLDO

        Parameters
        ----------
        gb_dict: dict
            dict containing details about the grain boundary

        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        #mark that the structure has a defect

        plane_defect_01 = BNode(name)
        self.add((self.material, CMSO.hasDefect, plane_defect_01))
        
        if gb_dict["GBType"] is None:
            self.add((plane_defect_01, RDF.type, PLDO.GrainBoundary))
        elif gb_dict["GBType"] == "Twist":
            self.add((plane_defect_01, RDF.type, PLDO.TwistBoundary))
        elif gb_dict["GBType"] == "Tilt":
            self.add((plane_defect_01, RDF.type, PLDO.TiltBoundary))
        elif gb_dict["GBType"] == "Symmetric Tilt":
            self.add((plane_defect_01, RDF.type, PLDO.SymmetricTiltBoundary))
        elif gb_dict["GBType"] == "Mixed":
            self.add((plane_defect_01, RDF.type, PLDO.MixedBoundary))
        self.add((plane_defect_01, PLDO.hasSigmaValue, Literal(gb_dict["sigma"], datatype=XSD.integer)))
        
        #now mark that the defect is GB
        #uname = None
        #if name is not None:
        #    uname = f'{name}GrainBoundaryPlane'
        #gb_plane_01 = BNode(uname)
        self.add((plane_defect_01, PLDO.hasGBPlane, Literal(gb_dict["GBPlane"], 
                                                             datatype=XSD.string)))
        #self.add((gb_plane_01, RDF.type, PLDO.GrainBoundaryPlane))
        #self.add((gb_plane_01, PLDO.hasMillerIndices, Literal(gb_dict["GBPlane"], 
        #                                                     datatype=XSD.string)))
        
        #uname = None
        #if name is not None:
        #    uname = f'{name}RotationAxis'
        #rotation_axis_01 = BNode(uname)
        self.add((plane_defect_01, PLDO.hasRotationAxis, Literal(gb_dict["RotationAxis"], 
                                                             datatype=XSD.string)))
        #self.add((rotation_axis_01, RDF.type, PLDO.RotationAxis))
        #self.add((rotation_axis_01, PLDO.hasComponentX, Literal(gb_dict["RotationAxis"][0], datatype=XSD.float)))
        #self.add((rotation_axis_01, PLDO.hasComponentY, Literal(gb_dict["RotationAxis"][1], datatype=XSD.float)))
        #self.add((rotation_axis_01, PLDO.hasComponentZ, Literal(gb_dict["RotationAxis"][2], datatype=XSD.float)))

        #uname = None
        #if name is not None:
        #    uname = f'{name}MisorientationAngle'
        #misorientation_angle_01 = BNode(uname)
        self.add((plane_defect_01, PLDO.hasMisorientationAngle, Literal(gb_dict["MisorientationAngle"], datatype=XSD.float)))
        #self.add((misorientation_angle_01, RDF.type, PLDO.MisorientationAngle))
        #self.add((misorientation_angle_01, PLDO.hasAngle, Literal(gb_dict["MisorientationAngle"], datatype=XSD.float)))    
    
    def add_vacancy(self, concentration, number=None, name=None):
        """
        Add Vacancy details which will be annotated by PODO

        Parameters
        ----------
        concentration: float
            vacancy concentration, value should be between 0-1

        name
            if provided, the name will be used instead of random identifier

        Returns
        -------
        """

        vacancy_01 = BNode(name)
        self.add((self.material, CMSO.hasDefect, vacancy_01))
        self.add((vacancy_01, RDF.type, PODO.Vacancy))
        self.add((self.simulation_cell, PODO.hasVacancyConcentration, Literal(concentration, datatype=XSD.float)))
        if number is not None:
            self.add((self.simulation_cell, PODO.hasNumberOfVacancies, Literal(number, datatype=XSD.integer)))
        #if vacancy is added, atoms have to be deleted from the existing record!
        #this is indeed a tricky item


    def add_calculated_quantity(self, propertyname, value, unit=None, sample=None):
        prop = BNode()
        if sample is None:
            sample = self.sample
        self.add((sample, CMSO.hasCalculatedProperty, prop))
        self.add((prop, RDF.type, CMSO.CalculatedProperty))
        self.add((prop, RDFS.label, Literal(propertyname)))
        self.add((prop, CMSO.hasValue, Literal(value)))
        if unit is not None:
            self.add((prop, CMSO.hasUnit, URIRef(f'https://qudt.org/2.1/vocab/unit#{unit}')))


    def inspect_sample(self, sample=None):
        if sample is None:
            sample = self.sample
        natoms = self.graph.value(sample, CMSO.hasNumberOfAtoms).toPython()
        material = list([k[2] for k in self.graph.triples((sample, CMSO.hasMaterial, None))])[0]
        defects = list([k[2] for k in self.graph.triples((material, CMSO.hasDefect, None))])
        composition = list([k[2].toPython() for k in self.graph.triples((material, CMSO.hasElementRatio, None))])
        crystalstructure = self.graph.value(material, CMSO.hasStructure)
        spacegroupsymbol = self.graph.value(crystalstructure, CMSO.hasSpaceGroupSymbol).toPython()

        lattice = self.graph.value(sample, CMSO.hasNumberOfAtoms).toPython()
        defect_types = list([self.graph.value(d, RDF.type).toPython() for d in defects])
        prop_nodes = list([k[2] for k in self.graph.triples((sample, CMSO.hasCalculatedProperty, None))])
        props = list([self.graph.value(prop_node, RDFS.label) for prop_node in prop_nodes])
        propvals = list([self.graph.value(d, CMSO.hasValue).toPython() for d in prop_nodes])
        units = list([self.graph.value(d, CMSO.hasUnit).toPython() for d in prop_nodes])
        st = []
        st.append(f'Sample with {natoms} atoms.\n')
        st.append("Material:\n")
        st.append(" ".join(composition))
        st.append("\n")
        st.append(f'Space Group symbol: {spacegroupsymbol}\n')
        if len(defect_types) > 0:
            st.append('With defects:\n')
            for d in defect_types:
                st.append(f'{d}\n')
        if len(props) > 0:
            st.append('With calculated properties:\n')
            for x in range(len(props)):
                st.append(f'{props[x]} with value: {propvals[x]} and unit: {units[x]}\n')

        return " ".join(st)

    def visualize(self, *args, **kwargs):
        """
        Vosualise the RDF tree of the Graph

        Parameters
        ----------
        backend: string, {'ipycytoscape', 'graphviz'}
            Chooses the backend with which the graph will be plotted. ipycytoscape provides an interactive, 
            but slow visualisation, whereas graphviz provides a non-interactive fast visualisation.

        edge_color: string
            Edge color of the boxes

        styledict: dict
            If provided, allows customisation of color and other properties.

        graph_attr: dict
            further attributes that allow customisation of graphs

        layoutname: string
            name of the layout for graph

        Returns
        -------

        Notes
        -----
        styledict has the following options. Refer to graphviz and ipycytoscape
        documentation for more details
        BNode:  
          color:  
          shape:  
          style:  
        URIRef:  
          color:  
          shape:  
          style:  
        Literal:  
          color:  
          shape:  
          style:         
        """
        self.visualise(*args, **kwargs)
        
    def visualise(self,
                  backend='ipycytoscape',
                  edge_color="#37474F",
                  styledict=None, 
                  graph_attr ={'rankdir': 'BT'},
                  layoutname='cola'):
        """
        Vosualise the RDF tree of the Graph

        Parameters
        ----------
        backend: string, {'ipycytoscape', 'graphviz'}
            Chooses the backend with which the graph will be plotted. ipycytoscape provides an interactive, 
            but slow visualisation, whereas graphviz provides a non-interactive fast visualisation.

        edge_color: string
            Edge color of the boxes

        styledict: dict
            If provided, allows customisation of color and other properties.

        graph_attr: dict
            further attributes that allow customisation of graphs

        layoutname: string
            name of the layout for graph

        Returns
        -------

        Notes
        -----
        styledict has the following options. Refer to graphviz and ipycytoscape
        documentation for more details
        BNode:  
          color:  
          shape:  
          style:  
        URIRef:  
          color:  
          shape:  
          style:  
        Literal:  
          color:  
          shape:  
          style:         
        """
        
        sdict = defstyledict.copy()
        if styledict is not None:
            sdict = _replace_keys(sdict, styledict)
        return visualize_graph(self.graph, 
                               backend=backend,
                               edge_color=edge_color,
                               styledict=sdict, 
                               graph_attr=graph_attr,
                               layoutname=layoutname)
    
    
    def write(self, filename, format="json-ld"):
        """
        Write the serialised version of the graph to a file

        Parameters
        ----------
        filename: string
            name of output file

        format: string, {'turtle', 'xml', 'json-ld', 'ntriples', 'n3'}
            output format to be written to 

        Returns
        -------
        None
        """

        with open(filename, "w") as fout:
            fout.write(self.graph.serialize(format=format))
            
        
    def to_file(self, sample, filename=None, format="lammps-dump"):
        """
        Save a given sample to a file

        Parameters
        ----------
        sample
            ID of the sample

        filename: string
            name of output file

        format: string, {"lammps-dump","lammps-data", "poscar"}

        Returns
        -------
        None
        """

        if filename is None:
            filename = os.path.join(os.getcwd(), "out")
        sys = self.get_system_from_sample(sample)
        if format=="ase":
            return sys.to_ase()
        elif format=='poscar':
            asesys = sys.to_ase()
            write(filename, asesys, format="vasp")
        else:
            #asesys = sys.to_ase()
            #write(filename, asesys, format=format)
            sys.to_file(filename, format=format)
    
        
    def query_sample(self, target_property, value, return_query=False):
        """
        Query the Graph for a sample that has the given `value` for the given `target_property`

        Parameters
        ----------
        target_property: string
            The target property can be any Ontology data property

        value: number, string, or list of either
            The value of the target property to be queried

        return_query: bool, optional
            If True, return the SPARQL query 

        Returns
        -------
        res: list
            list of queried samples

        query: string
            only returned if `return_query` is True

        Notes
        -----
        """

        query = self._query_graph.formulate_query(target_property, value)
        res = self.graph.query(query)
        res = [r for r in res]
        if return_query:
            return res, query
        return res
    
    #################################
    # Methods to interact with sample
    #################################
    @property
    def n_samples(self):
        """
        Number of samples in the Graph
        """

        return len([x for x in self.graph.triples((None, RDF.type, CMSO.AtomicScaleSample))])
    
    @property
    def samples(self):
        """
        Returns a list of all Samples in the graph
        """

        return [x[0] for x in self.graph.triples((None, RDF.type, CMSO.AtomicScaleSample))]
        
    def iterate_graph(self, item, create_new_graph=False):
        if create_new_graph:
            self.sgraph = RDFGraph()
        triples = list(self.graph.triples((item, None, None)))
        for triple in triples:
            self.sgraph.graph.add(triple)
            self.iterate_graph(triple[2])
    
    def get_sample(self, sample, no_atoms=False):
        """
        Get the Sample as an RDFGraph

        Parameters
        ----------
        sample: string
            sample id

        no_atoms: bool, optional
            if True, returns the number of atoms in the sample

        Returns
        -------
        sgraph: :py:class:`RDFGraph`
            the RDFGraph of the queried sample

        """

        self.iterate_graph(sample, create_new_graph=True)
        if no_atoms:
            na = self.sgraph.graph.value(sample, CMSO.hasNumberOfAtoms).toPython()
            return self.sgraph, na
        return self.sgraph
        
    def get_system_from_sample(self, sample):
        """
        Get a pyscal :py:class:`pyscal.core.System` from the selected sample

        Parameters
        ----------
        sample: string
            sample id

        Returns
        -------
        system: :py:class:`pyscal.core.System`
            corresponding system
        """

        simcell = self.graph.value(sample, CMSO.hasSimulationCell)
        cell_vectors = [[], [], []]

        for s in self.graph.triples((simcell, CMSO.hasVector, None)):
            cell_vectors[0].append(self.graph.value(s[2], CMSO.hasComponent_x).toPython())
            cell_vectors[1].append(self.graph.value(s[2], CMSO.hasComponent_y).toPython())
            cell_vectors[2].append(self.graph.value(s[2], CMSO.hasComponent_z).toPython())
        #cell_vectors

        positions = []
        species = []

        for atom in self.graph.triples((sample, CMSO.hasAtom, None)):
            vector = self.graph.value(atom[2], CMSO.hasPositionVector)
            pt = []
            pt.append(self.graph.value(vector, CMSO.hasComponent_x).toPython())
            pt.append(self.graph.value(vector, CMSO.hasComponent_y).toPython())
            pt.append(self.graph.value(vector, CMSO.hasComponent_z).toPython())
            element = self.graph.value(atom[2], CMSO.hasElement)
            species.append(self.graph.value(element, CMSO.hasSymbol).toPython())
            positions.append(pt)
        
        atoms = {"positions": positions, "species": species}
        at = Atoms()
        at.from_dict(atoms)
        sys = System()
        sys.box = cell_vectors
        sys.atoms = at
        
        return sys
